@@ -12,6 +12,9 @@
 #' @param tweak An optional function that takes an expression
 #'        and returns a tweaked expression.
 #'
+#' @param locals Should globals part of any "local" environment of
+#'        a function be included or not?
+#'
 #' @param substitute If TRUE, the expression is \code{substitute()}:ed,
 #'        otherwise not.
 #'
@@ -64,20 +67,25 @@
 #' @export
 globalsOf <- function(expr, envir = parent.frame(), ...,
                       method = c("ordered", "conservative", "liberal"),
-                      tweak = NULL, substitute = FALSE, mustExist = TRUE,
+                      tweak = NULL,
+                      locals = NA,
+                      substitute = FALSE, mustExist = TRUE,
                       unlist = TRUE, recursive = TRUE, skip = NULL) {
   method <- match.arg(method, choices = c("ordered", "conservative", "liberal"))
 
+  if (is.na(locals)) locals <- getOption("globals.globalsOf.locals", TRUE)
+  stop_if_not(is.logical(locals), length(locals) == 1L, !is.na(locals))
+  
   if (substitute) expr <- substitute(expr)
   stop_if_not(is.null(skip) || is.list(skip))
   
-  mdebug("globalsOf(..., method = '%s', mustExist = %s, unlist = %s, recursive = %s) ...", method, mustExist, unlist, recursive) #nolint
+  debug <- mdebug("globalsOf(..., method = '%s', mustExist = %s, unlist = %s, recursive = %s) ...", method, mustExist, unlist, recursive) #nolint
 
   ## 1. Identify global variables (static code inspection)
   names <- findGlobals(expr, envir = envir, ..., method = method,
                        tweak = tweak, substitute = FALSE, unlist = unlist)
-  mdebug(" - preliminary globals (by name): [%d] %s",
-         length(names), hpaste(sQuote(names)))
+  debug && mdebug(" - preliminary globals (by name): [%d] %s",
+                  length(names), hpaste(sQuote(names)))
 
   ## 2. Locate them (run time)
   globals <- tryCatch({
@@ -90,71 +98,92 @@ globalsOf <- function(expr, envir = parent.frame(), ...,
     stop(ex)
   })
 
-  mdebug(" - preliminary globals (by value): [%d] %s",
-         length(globals), hpaste(sQuote(names(globals))))
+  debug && mdebug(" - preliminary globals (by value): [%d] %s",
+                  length(globals), hpaste(sQuote(names(globals))))
+
+  ## If a function, drop any globals that are part of any of the functions
+  ## local environments, e.g. 'a' in f <- local({ a <- 1; function() a })
+  if (!locals && is.function(expr) && length(globals) > 0) {
+    env <- environment(expr) ## the environment of the function
+    eenv <- emptyenv()
+    genv <- globalenv()
+    where <- attr(globals, "where", exact = TRUE)
+    while (length(where) > 0 && !identical(env, eenv) && !identical(env, genv)) {
+      ## Any 'where' for the current environment?
+      keep <- !vapply(where, FUN.VALUE = FALSE, FUN = identical, env)
+      where <- where[keep]
+      env <- parent.env(env)
+    }
+    ## Anything to drop?
+    if (length(where) != length(globals)) globals <- globals[names(where)]
+  }
 
   ## 3. Among globals that are closures (functions) and that exist outside
   ##    of namespaces ("packages"), check for additional globals?
   if (recursive) {
-    mdebug(" - recursive scan of preliminary globals ...")
+    debug && mdebug(" - recursive scan of preliminary globals ...")
 
     ## Don't enter functions in namespaces / packages
-    where <- attr(globals, "where")
+    where <- attr(globals, "where", exact = TRUE)
     stop_if_not(length(where) == length(globals))
-    where <- sapply(where, FUN = envname)
+    where <- vapply(where, FUN = envname, FUN.VALUE = NA_character_,
+                    USE.NAMES = FALSE)
     globals_t <- globals[!(where %in% loadedNamespaces())]
 
-    mdebug(" - subset of globals to be scanned (not in loaded namespaces): [%d] %s", length(globals_t), hpaste(sQuote(names(globals_t)))) #nolint
+    debug && mdebug(" - subset of globals to be scanned (not in loaded namespaces): [%d] %s", length(globals_t), hpaste(sQuote(names(globals_t)))) #nolint
 
     ## Enter only functions
     ## NOTE: This excludes functions "not found", but also primitives
     ##       not dropped above.
-    globals_t <- globals_t[sapply(globals_t, FUN = typeof) == "closure"]
+    globals_t <- globals_t[vapply(globals_t, FUN = typeof, FUN.VALUE = NA_character_, USE.NAMES = FALSE) == "closure"]
 
     if (length(globals_t) > 0) {
-      mdebug(" - subset of globals to be scanned: [%d] %s",
-             length(globals_t), hpaste(sQuote(names(globals_t))))
+      debug && mdebug(" - subset of globals to be scanned: [%d] %s",
+                      length(globals_t), hpaste(sQuote(names(globals_t))))
       names_t <- names(globals_t)
 
       ## Avoid recursive scanning of already scanned ("known") globals
       skip_t <- c(skip, globals_t)
       
       for (gg in seq_along(globals_t)) {
-        mdebug("   + scanning global #%d (%s) ...", gg, sQuote(names_t[[gg]]))
+        debug && mdebug("   + scanning global #%d (%s) ...",
+	                gg, sQuote(names_t[[gg]]))
         fcn <- globals_t[[gg]]
 
         ## Is function 'fcn' among the already identified globals?
-        already_scanned <- any(sapply(skip, FUN = identical, fcn))
+        already_scanned <- any(vapply(skip, FUN = identical, fcn, FUN.VALUE = NA, USE.NAMES = FALSE))
         if (already_scanned) next;
 
         env <- environment(fcn) ## was 'env <- envir' in globals 0.8.0.
 
         globals_gg <- globalsOf(fcn, envir = env, ..., method = method,
-                                tweak = tweak, substitute = FALSE,
+                                tweak = tweak,
+                                locals = locals,
+                                substitute = FALSE,
                                 mustExist = mustExist, unlist = unlist,
                                 recursive = recursive,
                                 skip = skip_t)
         if (length(globals_gg) > 0) {
           globals <- c(globals, globals_gg)
 
-          skip_gg <- globals_gg[sapply(globals_gg, FUN = typeof) == "closure"]
+          skip_gg <- globals_gg[vapply(globals_gg, FUN = typeof, FUN.VALUE = NA_character_, USE.NAMES = FALSE) == "closure"]
           skip_t <- c(skip_t, skip_gg)
         }
       }
       globals <- unique(globals)
-      mdebug(" - updated set of globals found: [%d] %s",
-             length(globals), hpaste(sQuote(names(globals))))
+      debug && mdebug(" - updated set of globals found: [%d] %s",
+                      length(globals), hpaste(sQuote(names(globals))))
     } else {
-      mdebug(" - subset of globals to be scanned: [0]")
+      debug && mdebug(" - subset of globals to be scanned: [0]")
     }
 
-    mdebug(" - recursive scan of preliminary globals ... DONE")
+    debug && mdebug(" - recursive scan of preliminary globals ... DONE")
   }
 
-  mdebug(" - globals found: [%d] %s",
-         length(globals), hpaste(sQuote(names(globals))))
+  debug && mdebug(" - globals found: [%d] %s",
+                  length(globals), hpaste(sQuote(names(globals))))
 
-  mdebug("globalsOf(..., method = '%s', mustExist = %s, unlist = %s, recursive = %s) ... DONE", method, mustExist, unlist, recursive) #nolint
+  debug && mdebug("globalsOf(..., method = '%s', mustExist = %s, unlist = %s, recursive = %s) ... DONE", method, mustExist, unlist, recursive) #nolint
 
   globals
 } ## globalsOf()
@@ -178,23 +207,31 @@ globalsByName <- function(names, envir = parent.frame(), mustExist = TRUE,
                           ...) {
   names <- as.character(names)
 
-  mdebug("globalsByName(<%d names>, mustExist = %s) ...",
-         length(names), mustExist)
-  mdebug("- search from environment: %s", sQuote(envname(envir)))
+  nnames <- length(names)
+  
+  debug <- mdebug("globalsByName(<%d names>, mustExist = %s) ...",
+                  nnames, mustExist)
+  debug && mdebug("- search from environment: %s", sQuote(envname(envir)))
 
   ## Locate and retrieve the specified globals
-  idxs <- which(names == "...")
-  needs_dotdotdot <- (length(idxs) > 0)
-  if (needs_dotdotdot) names <- names[-idxs]
-  mdebug("- dotdotdot: %s", needs_dotdotdot)
+  idxs <- grep("^[.][.]([.]|[0-9]+)$", names)
+  if (length(idxs) > 0L) {
+    dotdotdots <- unique(names[idxs])
+    names <- names[-idxs]
+    idxs <- NULL
+    debug && mdebug("- dotdotdots: %s", paste(sQuote(dotdotdots), collapse = ", "))
+  } else {
+    dotdotdots <- NULL
+    debug && mdebug("- dotdotdots: <none>")
+  }
 
   globals <- structure(list(), class = c("Globals", "list"))
   where <- list()
   for (kk in seq_along(names)) {
     name <- names[kk]
-    mdebug("- locating #%d (%s)", kk, sQuote(name))
+    debug && mdebug("- locating #%d (%s)", kk, sQuote(name))
     env <- where(name, envir = envir, inherits = TRUE)
-    mdebug("  + found in environment: %s", sQuote(envname(env)))
+    debug && mdebug("  + found in environment: %s", sQuote(envname(env)))
     if (!is.null(env)) {
       where[[name]] <- env
       value <- get(name, envir = env, inherits = FALSE)
@@ -212,16 +249,19 @@ globalsByName <- function(names, envir = parent.frame(), mustExist = TRUE,
     }
   }
 
-  if (needs_dotdotdot) {
-    if (exists("...", envir = envir, inherits = TRUE)) {
-      where[["..."]] <- where("...", envir = envir, inherits = TRUE)
-      ddd <- evalq(list(...), envir = envir, enclos = envir)
-    } else {
-      where["..."] <- list(NULL)
-      ddd <- NA
+  if (length(dotdotdots) > 0L) {
+    for (name in dotdotdots) {
+      if (exists(name, envir = envir, inherits = TRUE)) {
+        where[[name]] <- where(name, envir = envir, inherits = TRUE)
+        expr <- substitute(list(arg), list(arg = as.name(name)))
+        ddd <- eval(expr, envir = envir, enclos = envir)
+      } else {
+        where[name] <- list(NULL)
+        ddd <- NA
+      }
+      class(ddd) <- c("DotDotDotList", class(ddd))
+      globals[[name]] <- ddd
     }
-    class(ddd) <- c("DotDotDotList", class(ddd))
-    globals[["..."]] <- ddd
   }
 
   stop_if_not(
@@ -232,8 +272,8 @@ globalsByName <- function(names, envir = parent.frame(), mustExist = TRUE,
 
   attr(globals, "where") <- where
 
-  mdebug("globalsByName(<%d names>, mustExist = %s) ... DONE",
-         length(names), mustExist)
+  debug && mdebug("globalsByName(<%d names>, mustExist = %s) ... DONE",
+                  nnames, mustExist)
 
   globals
 } ## globalsByName()
