@@ -1,5 +1,6 @@
 as_function <- function(expr, envir = parent.frame(), enclos = baseenv(), ...) {
-  eval(substitute(function() x, list(x = expr)), envir = envir, enclos = enclos, ...)
+  fun_expr <- substitute(function() x, list(x = expr))
+  eval(fun_expr, envir = envir, enclos = enclos, ...)
 }
 
 # Although the set of "base" packages rarely changes, it has happened
@@ -127,6 +128,11 @@ hexpr <- function(expr, trim = TRUE, collapse = "; ", max_head = 6L,
 } # hexpr()
 
 
+now <- function(x = Sys.time(), format = "[%H:%M:%OS3] ") {
+  ## format(x, format = format) ## slower
+  format(as.POSIXlt(x, tz = ""), format = format)
+}
+
 ## From future 1.3.0
 mdebug <- function(...) {
   if (!getOption("globals.debug", FALSE)) return(invisible(FALSE))
@@ -134,6 +140,12 @@ mdebug <- function(...) {
   invisible(TRUE)
 } ## mdebug()
 
+
+mdebugf <- function(..., appendLF = TRUE, prefix = now(), debug = getOption("globals.debug", FALSE)) {
+  if (!debug) return()
+  message(prefix, sprintf(...), appendLF = appendLF)
+}
+
 #' @importFrom utils capture.output str
 mstr <- function(...) {
   bfr <- capture.output(str(...))
@@ -170,6 +182,8 @@ envname <- function(env) {
   name
 }
 
+commaq <- function(x, sep = ", ") paste(sQuote(x), collapse = sep)
+
 ## When 'default' is specified, this is 30x faster than
 ## base::getOption().  The difference is that here we use
 ## use names(.Options) whereas in 'base' names(options())
@@ -181,6 +195,14 @@ getOption <- local({
   }
 })
 
+stopf <- function(fmt, ..., call. = TRUE, domain = NULL) {  #nolint
+  stop(sprintf(fmt, ...), call. = call., domain = domain)
+}
+
+warnf <- function(fmt, ..., call. = TRUE, domain = NULL) {  #nolint
+  warning(sprintf(fmt, ...), call. = call., domain = domain)
+}
+
 stop_if_not <- function(...) {
   res <- list(...)
   n <- length(res)
@@ -245,3 +267,52 @@ list_apply <- function(X, FUN, ...) {
   }
   res
 }
+
+
+.trace <- new.env()
+.trace$indent <- 0L
+
+trace_indent <- function(x = "", indent = .trace$indent) {
+#  utils::str(list(indent = indent))
+#  indent <- max(0L, indent)
+  prefix <- paste(rep(" ", times = 3*indent), collapse = "")
+  paste(prefix, x, sep = "")
+}
+
+trace_printf <- function(..., indent = .trace$indent, collapse = "\n", appendLF = FALSE) {
+  msg <- sprintf(...)
+  out <- trace_indent(msg, indent = indent)
+  out <- paste(out, collapse = collapse)
+  message(out, appendLF = appendLF)
+  invisible(msg)
+}
+
+#' @importFrom utils capture.output
+trace_print <- function(..., envir = parent.frame(), indent = .trace$indent, collapse = "\n", appendLF = TRUE) {
+  bfr <- eval(capture.output(print(...)), envir = envir)
+  trace_printf(bfr, indent = indent, collapse = collapse, appendLF = appendLF)
+}
+
+#' @importFrom utils capture.output str
+trace_str <- function(..., envir = parent.frame(), indent = .trace$indent, collapse = "\n", appendLF = TRUE) {
+  bfr <- eval(capture.output(str(...)), envir = envir)
+  trace_printf(bfr, indent = indent, collapse = collapse, appendLF = appendLF)
+}
+
+trace_enter <- function(..., appendLF = TRUE) {
+  msg <- trace_printf(..., appendLF = FALSE)
+  message(" ...", appendLF = appendLF)
+  .trace$indent <- .trace$indent + 1L
+  attr(msg, "indent") <- .trace$indent
+  invisible(msg)
+}
+
+trace_exit <- function(fmtstr, ..., appendLF = TRUE) {
+  indent <- attr(fmtstr, "indent")
+  if (!is.null(indent)) .trace$indent <- indent
+  .trace$indent <- .trace$indent - 1L
+  msg <- trace_printf(fmtstr, ..., appendLF = FALSE)
+  message(" ... done", appendLF = appendLF)
+#  stop_if_not(.trace$indent >= 0L)
+  invisible(msg)
+}
